'use strict'

var shimmer = require('../../shimmer')
var logger = require('../../logger').child({component: 'http'})
var recordWeb = require('../../metrics/recorders/http')
var hashes = require('../../util/hashes')
var cat = require('../../util/cat')
var instrumentOutbound = require('./http-outbound')
var util = require('util')
var url = require('url')
var urltils = require('../../util/urltils')
var properties = require('../../util/properties')
var psemver = require('../../util/process-version')
var copy = require('../../util/copy')

var NAMES = require('../../metrics/names')

/*
 *
 * CONSTANTS
 *
 */
var SHOULD_WRAP_HTTPS = psemver.satisfies('<0.11 || >=9.0.0 || 8.9.0')
var NR_CONNECTION_PROP = '__NR__connection'
var DEFAULT_HOST = 'localhost'
var DEFAULT_PORT = 80
var REQUEST_HEADER = 'x-request-start'
var QUEUE_HEADER = 'x-queue-start'
var NEWRELIC_ID_HEADER = 'x-newrelic-id'
var NEWRELIC_APP_DATA_HEADER = 'x-newrelic-app-data'
var NEWRELIC_TRANSACTION_HEADER = 'x-newrelic-transaction'
var NEWRELIC_SYNTHETICS_HEADER = 'x-newrelic-synthetics'
var CONTENT_LENGTH_REGEX = /^Content-Length$/i
var TRANSACTION_INFO_KEY = '__NR_transactionInfo'
var COLLECTED_REQUEST_HEADERS = [
  'accept',
  'content-length',
  'content-type',
  'referer',
  'user-agent',
  'host'
]
var HEADER_ATTR_NAMES = {
  'accept':                       'accept',
  'accept-charset':               'acceptCharset',
  'accept-encoding':              'acceptEncoding',
  'access-control-allow-headers': 'accessControlAllowHeaders',
  'access-control-allow-methods': 'accessControlAllowMethods',
  'access-control-allow-origin':  'accessControlAllowOrigin',
  'age':                          'age',
  'allow':                        'allow',
  'authorization':                'authorization',
  'cache-control':                'cacheControl',
  'connection':                   'connection',
  'cookie':                       'cookie',
  'content-encoding':             'contentEncoding',
  'content-length':               'contentLength',
  'content-type':                 'contentType',
  'date':                         'date',
  'etag':                         'eTag',
  'expect':                       'expect',
  'expires':                      'expires',
  'forwarded':                    'forwarded',
  'host':                         'host',
  'if-match':                     'ifMatch',
  'if-modified-since':            'ifModifiedSince',
  'last-modified':                'lastModified',
  'location':                     'location',
  'origin':                       'origin',
  'proxy-authorization':          'proxyAuthorization',
  'referer':                      'referer',
  'refresh':                      'refresh',
  'server':                       'server',
  'set-cookie':                   'setCookie',
  'transfer-encoding':            'transferEncoding',
  'user-agent':                   'userAgent',
  'upgrade':                      'upgrade',
  'vary':                         'vary',
  'x-correlation-id':             'xCorrelationId',
  'x-csrf-token':                 'xCsrfToken',
  'x-forwarded-for':              'xForwardedFor',
  'x-http-method-override':       'xHttpMethodOverride',
  'x-newrelic-app-data':          'xNewrelicAppData',
  'x-newrelic-id':                'xNewrelicId',
  'x-newrelic-synthetics':        'xNewrelicSynthetics',
  'x-newrelic-transaction':       'xNewrelicTransaction',
  'x-powered-by':                 'xPoweredBy',
  'x-queue-start':                'xQueueStart',
  'x-request-id':                 'xRequestId',
  'x-request-start':              'xRequestStart',
  'x-requested-with':             'xRequestedWith'
}
var REQUEST_HEADER_PREFIX = 'request.headers.'
var RESPONSE_HEADER_PREFIX = 'response.headers.'
var REQUEST_HEADER_NAMES = Object.create(null)
var RESPONSE_HEADER_NAMES = Object.create(null)
_setHeaderAttrNames(REQUEST_HEADER_NAMES, REQUEST_HEADER_PREFIX)
_setHeaderAttrNames(RESPONSE_HEADER_NAMES, RESPONSE_HEADER_PREFIX)


// For incoming requests this instrumentation functions by wrapping
// `http.createServer` and `http.Server#addListener`. The former merely sets the
// agent dispatcher to 'http' and the latter wraps any event handlers bound to
// `request`.
//
// The `request` event listener wrapper creates a transaction proxy which will
// start a new transaction whenever a new request comes in. It also scans the
// headers of the incoming request looking for CAT and synthetics headers.

function wrapEmitWithTransaction(agent, emit) {
  var tracer = agent.tracer
  var serverPort = null

  return tracer.transactionProxy(function wrappedHandler(evnt, request, response) {
    var transaction = tracer.getTransaction()
    if (!transaction) return emit.apply(this, arguments)

    transaction.nameState.setPrefix(NAMES.NODEJS.PREFIX)
    transaction.nameState.setDelimiter(NAMES.ACTION_DELIMITER)

    // Store the transaction information on the request and response.
    storeTxInfo(transaction, request, response)

    // Hook for web framework instrumentations that don't have easy access to
    // the request entry point.
    if (properties.hasOwn(this, '__NR_onRequestStarted')) {
      this.__NR_onRequestStarted(request, response)
    }

    if (request) {
      initializeRequest(transaction, request)
    }

    // Create the transaction segment using the request URL for now. Once a
    // better name can be determined this segment will be renamed to that.
    var segment = tracer.createSegment(request.url, recordWeb)
    segment.start()

    if (agent.config.feature_flag.custom_instrumentation) {
      transaction.type = 'web'
      transaction.baseSegment = segment
    }

    /* Needed for Connect and Express middleware that monkeypatch request
     * and response via listeners.
     */
    tracer.bindEmitter(request, segment)
    tracer.bindEmitter(response, segment)

    // the error tracer needs a URL for tracing, even though naming overwrites
    transaction.parsedUrl = url.parse(request.url, true)
    transaction.url = transaction.parsedUrl.pathname
    transaction.verb = request.method

    // URL is sent as an agent attribute with transaction events
    transaction.addAgentAttribute('request.uri', transaction.url)

    // store the port on which this transaction runs
    if (this.address instanceof Function) {
      var address = this.address()
      if (address) {
        serverPort = address.port
      }
    }
    transaction.port = serverPort

    // need to set any config-driven names early for RUM
    logger.trace({url: request.url, transaction: transaction.id},
      'Applying user naming rules for RUM.')
    transaction.applyUserNamingRules(request.url)

    /**
     * Calculate Queue Time
     *
     * Queue time is provided by certain providers by stamping the request
     * header with the time the request arrived at the router.
     *
     * Units for queue time are
     */
    var qtime = request.headers[REQUEST_HEADER] || request.headers[QUEUE_HEADER]
    if (qtime) {
      var split = qtime.split('=')
      if (split.length > 1) {
        qtime = split[1]
      }

      var start = parseFloat(qtime)

      if (isNaN(start)) {
        logger.warn('Queue time header parsed as NaN (%s)', qtime)
      } else {
        // nano seconds
        if (start > 1e18) start = start / 1e6
        // micro seconds
        else if (start > 1e15) start = start / 1e3
        // seconds
        else if (start < 1e12) start = start * 1e3

        transaction.queueTime = Date.now() - start
      }
    }
    if (agent.config.cross_application_tracer.enabled) {
      var encKey = agent.config.encoding_key
      var incomingCatId = request.headers[NEWRELIC_ID_HEADER]
      var obfTransaction = request.headers[NEWRELIC_TRANSACTION_HEADER]
      var synthHeader = request.headers[NEWRELIC_SYNTHETICS_HEADER]
      if (encKey) {
        cat.handleCatHeaders(incomingCatId, obfTransaction, encKey, transaction)
        if (transaction.incomingCatId) {
          logger.trace('Got inbound request CAT headers in transaction %s',
            transaction.id)
        }
        if (synthHeader && agent.config.trusted_account_ids) {
          handleSyntheticsHeader(
            synthHeader,
            encKey,
            agent.config.trusted_account_ids,
            transaction
          )
        }
      }
    }

    function instrumentedFinish() {
      // Remove listeners so this doesn't get called twice.
      response.removeListener('finish', instrumentedFinish)
      request.removeListener('aborted', instrumentedFinish)

      // Naming must happen before the segment and transaction are ended,
      // because metrics recording depends on naming's side effects.
      transaction.finalizeNameFromUri(transaction.parsedUrl, response.statusCode)

      if (response) {
        if (response.statusCode != null) {
          var responseCode = String(response.statusCode)
          transaction.addAgentAttribute('httpResponseCode', responseCode)

          if (/^\d+$/.test(responseCode)) {
            transaction.addAgentAttribute('response.status', responseCode)
          }
        }
        if (response.statusMessage !== undefined) {
          transaction.addAgentAttribute('httpResponseMessage', response.statusMessage)
        }

        var headers = (response.getHeaders && response.getHeaders()) || response._headers
        if (headers) {
          _collectHeaders(
            headers,
            RESPONSE_HEADER_NAMES,
            RESPONSE_HEADER_PREFIX,
            transaction
          )
        }
      }

      // And we are done! End the segment and transaction.
      segment.end()
      transaction.end()
    }
    response.once('finish', instrumentedFinish)
    request.once('aborted', instrumentedFinish)

    return tracer.bindFunction(emit, segment).apply(this, arguments)
  })
}

function storeTxInfo(transaction, request, response) {
  if (!request || !response) {
    logger.debug('Missing request or response object! Not storing transaction info.')
    return
  }
  var hideInternal = transaction.agent.config.transaction_tracer.hide_internals

  var txInfo = {
    transaction: transaction,
    segmentStack: [],
    errorHandled: false,
    error: null
  }
  if (hideInternal) {
    properties.setInternal(request, TRANSACTION_INFO_KEY, txInfo)
    properties.setInternal(response, TRANSACTION_INFO_KEY, txInfo)
  } else {
    request[TRANSACTION_INFO_KEY] = response[TRANSACTION_INFO_KEY] = txInfo
  }

  logger.trace(
    'Stored transaction %s information on request and response',
    transaction.id
  )
}

function initializeRequest(transaction, request) {
  _collectHeaders(
    request.headers,
    REQUEST_HEADER_NAMES,
    REQUEST_HEADER_PREFIX,
    transaction
  )

  if (request.method != null) {
    transaction.addAgentAttribute('request.method', request.method)
    transaction.nameState.setVerb(request.method)
  }
}

function wrapResponseEnd(agent, proto) {
  var tracer = agent.tracer

  // On end, we must freeze the current name state to maintain the route that
  // responded and also end the current segment (otherwise it may become truncated).
  shimmer.wrapMethod(proto, 'Response.prototype', 'end', function wrapResEnd(end) {
    if (typeof end !== 'function') {
      logger.debug('Response#end is not a function?')
      return end
    }

    return function wrappedResEnd() {
      var txInfo = this && this[TRANSACTION_INFO_KEY]
      if (!txInfo) {
        return end.apply(this, arguments)
      }

      // If an error happend, add it to the aggregator.
      if (txInfo.error) {
        if (!txInfo.errorHandled || urltils.isError(agent.config, this.statusCode)) {
          agent.errors.add(txInfo.transaction, txInfo.error)
        }
      }

      // End all the segments leading up to and including this one.
      for (var i = txInfo.segmentStack.length - 1; i >= 0; --i) {
        txInfo.segmentStack[i].end()
      }
      var segment = tracer.getSegment()
      if (segment) {
        segment.end()
      }

      // Freeze the name state to prevent further changes.
      txInfo.transaction.nameState.freeze()

      return end.apply(this, arguments)
    }
  })
}

// CAT this wont be used unless CAT is enabled, see below where we actually do
// the shimmer stuff if you'd like to verify.
function wrapWriteHead(agent, writeHead) {
  return function wrappedWriteHead() {
    var transaction = agent.tracer.getTransaction()
    if (!transaction) {
      logger.trace('No transaction - not adding response CAT headers')
      return writeHead.apply(this, arguments)
    }
    // FLAG: synthetics
    if (agent.config.feature_flag.synthetics && transaction.syntheticsHeader) {
      this.setHeader(NEWRELIC_SYNTHETICS_HEADER, transaction.syntheticsHeader)
    }

    if (!transaction.incomingCatId) {
      logger.trace('No incoming CAT ID - not adding response CAT headers')
      return writeHead.apply(this, arguments)
    }

    if (!agent.config.trusted_account_ids) {
      logger.trace('No account IDs defined in config.trusted_account_ids - ' +
        'not adding response CAT headers')
      return writeHead.apply(this, arguments)
    }

    var accountId = transaction.incomingCatId.split('#')[0]
    accountId = parseInt(accountId, 10)
    if (agent.config.trusted_account_ids.indexOf(accountId) === -1) {
      logger.trace('Request from untrusted CAT header account id: %s - ' +
        'not adding response CAT headers', accountId)
      return writeHead.apply(this, arguments)
    }

    // Not sure this could ever happen, but should guard against it anyway
    // otherwise exception we blow up the user's app.
    if (!agent.config.cross_process_id || !agent.config.encoding_key) {
      logger.trace(
        'Managed to have agent.config.trusted_account_ids but not cross_process_id ' +
          '(%s) or encoding_key (%s) - not adding response CAT headers',
        agent.config.cross_process_id,
        agent.config.encoding_key
      )
      return writeHead.apply(this, arguments)
    }

    // -1 means no content length header was sent. We should only send this
    // value in the appData if the header is set.
    var contentLength = -1
    var new_headers = arguments[arguments.length - 1]

    if (typeof new_headers === 'object') {
      for (var header in new_headers) {  // jshint ignore: line
        if (CONTENT_LENGTH_REGEX.test(header)) {
          contentLength = new_headers[header]
          break
        }
      }
    }

    if (contentLength === -1 && this._headers) {
      // JSHint complains about ownProperty stuff, but since we are looking
      // for a specific name that doesn't matter so I'm disabling it.
      // Outbound headers can be capitalized in any way, use regex instead
      // of direct lookup.
      for (var userHeader in this._headers) {  // jshint ignore: line
        if (CONTENT_LENGTH_REGEX.test(userHeader)) {
          contentLength = this._headers[userHeader]
          break
        }
      }
    }
    // Stored on the tx so we can push a metric with this time instead of
    // actual duration.
    transaction.catResponseTime = transaction.timer.getDurationInMillis()

    var appData = null
    var txName = transaction.getFullName() || ''

    try {
      appData = JSON.stringify([
        agent.config.cross_process_id, // cross_process_id
        txName, // transaction name
        transaction.queueTime / 1000, // queue time (s)
        transaction.catResponseTime / 1000, // response time (s)
        contentLength, // content length (if content-length header is also being sent)
        transaction.id, // TransactionGuid
        false // force a transaction trace to be recorded
      ])
    } catch (err) {
      logger.trace(err, 'Failed to serialize transaction: %s - ' +
          'not adding CAT response headers',
        txName)
      return writeHead.apply(this, arguments)
    }

    var encKey = agent.config.encoding_key
    var obfAppData = hashes.obfuscateNameUsingKey(appData, encKey)
    this.setHeader(NEWRELIC_APP_DATA_HEADER, obfAppData)
    logger.trace('Added outbound response CAT headers in transaction %s', transaction.id)

    return writeHead.apply(this, arguments)
  }
}

function wrapRequest(agent, request) {
  // TODO: early return in the !transaction || internalOnly case
  return function wrappedRequest(options) {
    var tracer = agent.tracer
    var transaction = tracer.getTransaction()
    var outboundHeaders = Object.create(null)
    var args = tracer.slice(arguments)
    var context = this

    // don't pollute metrics and calls with NR connections
    var internalOnly = options && options[NR_CONNECTION_PROP]

    if (internalOnly) options[NR_CONNECTION_PROP] = undefined

    if (transaction && !internalOnly && agent.config.encoding_key) {
      // FLAG: synthetics
      if (agent.config.feature_flag.synthetics && transaction.syntheticsHeader) {
        outboundHeaders[NEWRELIC_SYNTHETICS_HEADER] = transaction.syntheticsHeader
      }

      // If CAT is enabled, inject the transaction header.
      if (agent.config.cross_application_tracer.enabled) {
        if (agent.config.obfuscatedId) {
          outboundHeaders[NEWRELIC_ID_HEADER] = agent.config.obfuscatedId
        }

        var pathHash = hashes.calculatePathHash(
          agent.config.applications()[0],
          transaction.getFullName() || '',
          transaction.referringPathHash
        )
        transaction.pushPathHash(pathHash)
        var txData = [
          transaction.id,
          false,
          transaction.tripId || transaction.id,
          pathHash
        ]
        try {
          txData = JSON.stringify(txData)
          var txHeader = hashes.obfuscateNameUsingKey(txData, agent.config.encoding_key)
          outboundHeaders[NEWRELIC_TRANSACTION_HEADER] = txHeader
          logger.trace('Added outbound request CAT headers in transaction %s',
            transaction.id)
        } catch (err) {
          logger.trace(err, 'Failed to serialize outbound response header')
        }
      }
    }

    if (transaction && !internalOnly) {
      if (typeof options === 'string') {
        options = url.parse(options)
      } else {
        options = copy.shallow(options)
      }

      if (util.isArray(options.headers)) {
        options.headers = options.headers.slice()
        Array.prototype.push.apply(
          options.headers,
          Object.keys(outboundHeaders).map(function getHeaderTuples(key) {
            return [key, outboundHeaders[key]]
          })
        )
      } else {
        options.headers = [options.headers, outboundHeaders]
          .reduce(function add(acc, val) {
            return copy.shallow(val, acc)
          }, Object.create(null))
      }
      args[0] = options

      // hostname & port logic pulled directly from node's 0.10 lib/http.js
      var hostname = options.hostname || options.host || DEFAULT_HOST
      var port = options.port || options.defaultPort || DEFAULT_PORT
      return instrumentOutbound(agent, hostname, port, function makeRequest() {
        return request.apply(context, args)
      })
    } else if (logger.traceEnabled() && !transaction && !internalOnly) {
      var logOpts = typeof options === 'string' ? url.parse(options) : options
      logger.trace(
        'No transaction, not recording external to %s:%s',
        logOpts.hostname || logOpts.host,
        logOpts.port
      )
    }

    return request.apply(context, args)
  }
}

function wrapLegacyRequest(agent, request) {
  return function wrappedLegacyRequest(method, path, headers) {
    var makeRequest = request.bind(this, method, path, headers)

    if (agent.tracer.getTransaction()) {
      return instrumentOutbound(agent, this.host, this.port, makeRequest)
    }

    logger.trace('No transaction, not recording external to %s:%s', this.host, this.port)
    return makeRequest()
  }
}

function wrapLegacyClient(agent, proto) {
  shimmer.wrapMethod(
    proto,
    'http.Client.prototype',
    'request',
    wrapLegacyRequest.bind(null, agent)
  )
}

module.exports = function initialize(agent, http, moduleName) {
  if (!http) {
    logger.debug('Did not get http module, not instrumenting!')
    return false
  }

  // FIXME: will this ever not be called?
  shimmer.wrapMethod(http, 'http', 'createServer', function cb_wrapMethod(createServer) {
    return function setDispatcher(requestListener) { // eslint-disable-line no-unused-vars
      agent.environment.setDispatcher('http')
      return createServer.apply(this, arguments)
    }
  })

  // It's not a great idea to monkeypatch EventEmitter methods given how hot
  // they are, but this method is simple and works with all versions of node
  // supported by the module.
  shimmer.wrapMethod(
    http.Server && http.Server.prototype,
    'http.Server.prototype',
    'emit',
    function wrapEmit(emit) {
      var txStarter = wrapEmitWithTransaction(agent, emit)
      return function wrappedEmit(evnt) {
        if (evnt === 'request') {
          return txStarter.apply(this, arguments)
        }
        return emit.apply(this, arguments)
      }
    }
  )

  wrapResponseEnd(agent, http.ServerResponse && http.ServerResponse.prototype)


  // If CAT is enabled we'll wrap `writeHead` to inject our headers.
  if (agent.config.cross_application_tracer.enabled) {
    shimmer.wrapMethod(
      http.ServerResponse && http.ServerResponse.prototype,
      'http.ServerResponse.prototype',
      'writeHead',
      wrapWriteHead.bind(null, agent)
    )
  }

  var agentProto = http && http.Agent && http.Agent.prototype

  // As of node 0.8, http.request() is the right way to originate outbound
  // requests. From 0.11 until 9, the `https` module simply called through to
  // the `http` methods, so to prevent double-instrumenting we need to check
  // what module we're instrumenting and what version of Node we're on. This
  // change originally also appeared in 8.9.0 but was reverted in 8.9.1
  if (SHOULD_WRAP_HTTPS || moduleName !== 'https') {
    shimmer.wrapMethod(
      http,
      'http',
      'request',
      wrapRequest.bind(null, agent)
    )

    if (psemver.satisfies('>=8')) {
      shimmer.wrapMethod(
        http,
        'http',
        'get',
        wrapRequest.bind(null, agent)
      )
    }
  }

  // Agent#createConnection was added in 0.11.
  // TODO: Remove this check when deprecating Node 0.10
  if (agentProto && agentProto.createConnection) {
    shimmer.wrapMethod(
      agentProto,
      'http.Agent.prototype',
      'createConnection',
      function wrapCreateConnection(original) {
        return function wrappedCreateConnection() {
          if (!agent.getTransaction()) {
            return original.apply(this, arguments)
          }

          var segment = agent.tracer.createSegment('http.Agent#createConnection')

          var args = agent.tracer.slice(arguments)
          if (typeof args[1] === 'function') {
            args[1] = agent.tracer.bindFunction(args[1], segment, true)
          }

          var sock = agent.tracer.bindFunction(original, segment, true)
            .apply(this, args)
          return sock
        }
      }
    )
  }

  // http.Client is deprecated, but still in use
  var DeprecatedClient, deprecatedCreateClient
  function clearGetters() {
    if (DeprecatedClient) {
      delete http.Client
      http.Client = DeprecatedClient
    }
    if (deprecatedCreateClient) {
      delete http.createClient
      http.createClient = deprecatedCreateClient
    }
  }

  DeprecatedClient = shimmer.wrapDeprecated(
    http,
    'http',
    'Client',
    {
      get: function get() {
        var example = new DeprecatedClient(80, 'localhost')
        wrapLegacyClient(agent, example.constructor.prototype)
        clearGetters()

        return DeprecatedClient
      },
      set: function set(NewClient) {
        DeprecatedClient = NewClient
      }
    }
  )

  deprecatedCreateClient = shimmer.wrapDeprecated(
    http,
    'http',
    'createClient',
    {
      get: function get() {
        var example = deprecatedCreateClient(80, 'localhost')
        wrapLegacyClient(agent, example.constructor.prototype)
        clearGetters()

        return deprecatedCreateClient
      },
      set: function set(newCreateClient) {
        deprecatedCreateClient = newCreateClient
      }
    }
  )
}

/**
 * Take the X-NewRelic-Synthetics header and apply any appropriate data to the
 * transaction for later use. This is the gate keeper for attributes being
 * added onto the transaction object for synthetics.
 *
 * @param {string} header - The raw X-NewRelic-Synthetics header
 * @param {string} encKey - Encoding key handed down from the server
 * @param {̄Array} trustedIds - Array of accounts to trust the header from.
 * @param {Transaction} transaction - Where the synthetics data is attached to.
 */
function handleSyntheticsHeader(header, encKey, trustedIds, transaction) {
  var synthData = parseSyntheticsHeader(header, encKey, trustedIds)
  if (!synthData) {
    return
  }

  transaction.syntheticsData = synthData
  transaction.syntheticsHeader = header
}

/**
 * Parse out and verify the the pieces of the X-NewRelic-Synthetics header.
 *
 * @param {string} header - The raw X-NewRelic-Synthetics header
 * @param {string} encKey - Encoding key handed down from the server
 * @param {̄Array} trustedIds - Array of accounts to trust the header from.
 * @return {Object or null} - On successful parse and verification an object of
 *                            synthetics data is returned, otherwise null is
 *                            returned.
 */
function parseSyntheticsHeader(header, encKey, trustedIds) {
  // Eagerly declare this object because we know what it should look like and
  // can use that for header verification.
  var parsedData = {
    version: null,
    accountId: null,
    resourceId: null,
    jobId: null,
    monitorId: null
  }
  var synthData = null
  try {
    synthData = JSON.parse(
      hashes.deobfuscateNameUsingKey(header, encKey)
    )
  } catch (e) {
    logger.trace(e, 'Got unparsable synthetics header: %s', header)
    return
  }

  if (!util.isArray(synthData)) {
    logger.trace(
      'Synthetics data is not an array: %s (%s)',
      synthData,
      typeof synthData
    )
    return
  }


  if (synthData.length < Object.keys(parsedData).length) {
    logger.trace(
      'Synthetics header length is %s, expected at least %s',
      synthData.length,
      Object.keys(parsedData).length
    )
  }

  parsedData.version = synthData[0]
  if (parsedData.version !== 1) {
    logger.trace(
      'Synthetics header version is not 1, got: %s (%s)',
      parsedData.version,
      synthData
    )
    return
  }

  parsedData.accountId = synthData[1]
  if (parsedData.accountId) {
    if (trustedIds.indexOf(parsedData.accountId) === -1) {
      logger.trace(
        'Synthetics header account ID is not in trusted account IDs: %s (%s)',
        parsedData.accountId,
        trustedIds
      )
      return
    }
  } else {
    logger.trace('Synthetics header account ID missing.')
    return
  }

  parsedData.resourceId = synthData[2]
  if (!parsedData.resourceId) {
    logger.trace('Synthetics resource ID is missing.')
    return
  }

  parsedData.jobId = synthData[3]
  if (!parsedData.jobId) {
    logger.trace('Synthetics job ID is missing.')
  }

  parsedData.monitorId = synthData[4]
  if (!parsedData.monitorId) {
    logger.trace('Synthetics monitor ID is missing.')
  }

  return parsedData
}

function _headerToCamelCase(header) {
  return header.replace(/[\W_]+(\w)/g, function capitalize(m, $1) {
    return $1.toUpperCase()
  })
}

function _setHeaderAttrNames(dest, prefix) {
  Object.keys(HEADER_ATTR_NAMES).forEach(function forEachHeader(h) {
    dest[h] = prefix + HEADER_ATTR_NAMES[h]
  })
}

function _collectHeaders(headers, nameMap, prefix, tx) {
  var headerKeys = !tx.agent.config.allow_all_headers
    ? COLLECTED_REQUEST_HEADERS
    : Object.keys(headers)

  for (var i = 0; i < headerKeys.length; i++) {
    var headerKey = headerKeys[i]
    var header = headers[headerKey]
    if (header !== undefined) {
      // If any more processing of the headers is required consider refactoring this.
      if (headerKey === 'referer' && typeof header === 'string') {
        var queryParamIndex = header.indexOf('?')
        if (queryParamIndex !== -1) {
          header = header.substring(0, queryParamIndex)
        }
      }

      var attributeName = nameMap[headerKey] || prefix + _headerToCamelCase(headerKey)
      tx.addAgentAttribute(attributeName, header)
    }
  }
}
